﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.BioBoost.Operators;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;

namespace HeuristicLab.BioBoost.Evaluators {

  [StorableClass]
  public abstract class BioBoostEvaluator : BioBoostOperator  {

    #region Parameter
    public ILookupParameter<ResultCollection> CostsParameter {
      get { return (ILookupParameter<ResultCollection>) Parameters["Costs"]; }
    }
    public ILookupParameter<ResultCollection> IntermediateResultsParameter {
      get { return (ILookupParameter<ResultCollection>) Parameters["IntermediateResults"]; }
    }
    #endregion

    #region Parameter Values
    public ResultCollection Costs {
      get { return CostsParameter.ActualValue; }
    }
    public ResultCollection IntermediateResults {
      get { return IntermediateResultsParameter.ActualValue; }
    }
    #endregion

    #region Construction & Cloning
    [StorableConstructor]
    protected BioBoostEvaluator(bool isDeserializing) : base(isDeserializing) {}
    protected BioBoostEvaluator(BioBoostEvaluator original, Cloner cloner) : base(original, cloner) {}
    protected BioBoostEvaluator() {
      Parameters.Add(new LookupParameter<ResultCollection>("Costs", "Collection of all costs."));
      Parameters.Add(new LookupParameter<ResultCollection>("Emissions", "Collection of all emissions."));
      Parameters.Add(new LookupParameter<ResultCollection>("IntermediateResults", "Collection of all intermeditate results."));
    }
    #endregion

    protected void PutInScope(string label, IItem value, bool isIntermediateVariable) {
      var vars = ExecutionContext.Scope.Variables;
      var var = new Variable(label, value);
      if (vars.ContainsKey(label))
        vars.Remove(label);
      vars.Add(var);
      if (isIntermediateVariable)
        IntermediateResults.Add(new Result(var.Name, value.GetType()));
    }

    protected void AddInScope(string label, DoubleArray values, bool isIntermediateValue) {
      var oldValues = GetFromScope<DoubleArray>(label);
      if (oldValues != null) {
        if (oldValues.Length != values.Length)
          throw new ArgumentException("DoubleArray lentgh differs from new length, cannot add " + label);
        for (int i = 0; i<oldValues.Length; i++)
          oldValues[i] += values[i];
      } else {
        PutInScope(label, values, isIntermediateValue);
      }
    }

    protected void AddCost(string label, double value) {
      if (Equals(value, 0d)) return;
      if (Costs.ContainsKey(label)) {
        ((DoubleValue)Costs[label].Value).Value += value;
      } else {
        Costs.Add(new Result(label, new DoubleValue(value)));
      }
    }
  }
}
