﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.BioBoost.Operators.Transformation;
using HeuristicLab.BioBoost.ProblemDescription;
using HeuristicLab.BioBoost.Utils;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;
using System.Linq;
using HeuristicLab.BioBoost.Representation;
using UniformOnePositionManipulator = HeuristicLab.Encodings.IntegerVectorEncoding.UniformOnePositionManipulator;
using UniformOnePositionRealVectorManipulator = HeuristicLab.Encodings.RealVectorEncoding.UniformOnePositionManipulator;

namespace HeuristicLab.BioBoost.Operators.Mutation {
  [StorableClass]
  public class CompoundMutator : SingleSuccessorOperator, IManipulator, IStochasticOperator {


    #region Parameters
    public ValueLookupParameter<IntMatrix> RegionBoundsParameter { get { return (ValueLookupParameter<IntMatrix>) Parameters["RegionBounds"]; } } 
    public LookupParameter<BioBoostProblemData> ProblemDataParameter { get { return (LookupParameter<BioBoostProblemData>) Parameters["ProblemData"]; } }
    public OptionalConstrainedValueParameter<IIntegerVectorManipulator> TransportTargetManipulatorParameter { get { return (OptionalConstrainedValueParameter<IIntegerVectorManipulator>) Parameters["TransportTargetManipulator"]; } }
    public OptionalConstrainedValueParameter<IRealVectorManipulator> FeedstockUtilizationManipulatorParameter { get { return (OptionalConstrainedValueParameter<IRealVectorManipulator>) Parameters["FeedstockUtilizationManipulator"]; } }
    public ILookupParameter<IRandom> RandomParameter { get { return (ILookupParameter<IRandom>) Parameters["Random"]; } }
    public ILookupParameter<IntValue> MaximumGenerationsParameter { get { return (ILookupParameter<IntValue>) Parameters["MaximumGenerations"]; } }
    public ILookupParameter<IntValue> GenerationsParameter { get { return (ILookupParameter<IntValue>) Parameters["Generations"]; } }
    public ILookupParameter<PercentValue> MinGenerationPercentageForUtilzationMutationParameter { get { return (ILookupParameter<PercentValue>) Parameters["MinGenerationPercentageForUtilzationMutation"]; } }
    public ILookupParameter<PercentValue> ProbabilityOfCombinedMutationParameter { get { return (ILookupParameter<PercentValue>) Parameters["ProbabilityOfCombinedMutation"]; } }
    #endregion

    #region Parameter Values
    public IIntegerVectorManipulator TransportTargetManipulator { get { return (IIntegerVectorManipulator) TransportTargetManipulatorParameter.ActualValue; } }
    public IRealVectorManipulator FeedstockUtilizationManipulator { get { return (IRealVectorManipulator) FeedstockUtilizationManipulatorParameter.ActualValue; } }
    public BioBoostProblemData ProblemData { get { return ProblemDataParameter.ActualValue; } }
    public IRandom Random { get { return RandomParameter.ActualValue; } }
    public int MaximumGenerations { get { return MaximumGenerationsParameter.ActualValue.Value; } }
    public int Generations { get { return GenerationsParameter.ActualValue.Value; } }
    public double MinGenerationPercentageForUtilzationMutation { get { return MinGenerationPercentageForUtilzationMutationParameter.ActualValue.Value; } }
    public double ProbabilityOfCombinedMutation { get { return ProbabilityOfCombinedMutationParameter.ActualValue.Value; } }
    #endregion

    #region Construction & Cloning
    [StorableConstructor]
    protected CompoundMutator(bool isDeserializing) : base(isDeserializing) {}
    protected CompoundMutator(CompoundMutator orig, Cloner cloner) : base(orig, cloner) { }
    public CompoundMutator() {
      Parameters.Add(new LookupParameter<BioBoostProblemData>("ProblemData", "The data store of the detailed problem description."));
      Parameters.Add(new ValueLookupParameter<IntMatrix>("RegionBounds", "The limits of valid region ids."));
      Parameters.Add(new OptionalConstrainedValueParameter<IIntegerVectorManipulator>("TransportTargetManipulator", "The manipulator to apply to the transport target vectors."));
      Parameters.Add(new OptionalConstrainedValueParameter<IRealVectorManipulator>("FeedstockUtilizationManipulator", "The manipulator for the feedstock utilization vectors."));
      Parameters.Add(new ValueLookupParameter<IRandom>("Random", "The random number generator used."));
      Parameters.Add(new LookupParameter<IntValue>("MaximumGenerations", "The maximum number of generations."));
      Parameters.Add(new LookupParameter<IntValue>("Generations", "The current number of generations."));
      Parameters.Add(new ValueLookupParameter<PercentValue>("MinGenerationPercentageForUtilzationMutation", "The minimum percentage of generations (Generations/MaxGenerations) until mutation of utilizations become enabled.", new PercentValue(0.5)));
      Parameters.Add(new ValueLookupParameter<PercentValue>("ProbabilityOfCombinedMutation", "Probability of mutating both utilization and transport. Don't set to 100% otherwise intermediates are not touched anymore.", new PercentValue(0.3)));
      foreach (var op in ApplicationManager.Manager.GetInstances<IRealVectorManipulator>()) {
        op.BoundsParameter.Value = new DoubleMatrix(new double[,] {{0, 1}});
        FeedstockUtilizationManipulatorParameter.ValidValues.Add(op);
      }
      foreach (var op in ApplicationManager.Manager.GetInstances<IIntegerVectorManipulator>()) {
        var bop = op as BoundedIntegerVectorManipulator;
        if (bop != null)
          bop.BoundsParameter.ActualName = RegionBoundsParameter.ActualName;
        if (op.Parameters.ContainsKey("Bounds") && op.Parameters["Bounds"] is ILookupParameter<IntMatrix>) {
          ((ILookupParameter<IntMatrix>) op.Parameters["Bounds"]).ActualName = RegionBoundsParameter.ActualName;
        }
        TransportTargetManipulatorParameter.ValidValues.Add(op);
      }
      FeedstockUtilizationManipulatorParameter.Value =
        FeedstockUtilizationManipulatorParameter.ValidValues.FirstOrDefault(
          c => c.GetType() == typeof (MultiRealVectorManipulator));
      TransportTargetManipulatorParameter.Value =
        TransportTargetManipulatorParameter.ValidValues.FirstOrDefault(
          c => c.GetType() == typeof (MultiIntegerVectorManipulator));
      var um = (MultiRealVectorManipulator) FeedstockUtilizationManipulator;
      foreach (var op in um.Operators) {
        if (op is PolynomialAllPositionManipulator) {
          ((PolynomialAllPositionManipulator) op).ContiguityParameter.Value = new DoubleValue(20);
        } else if (op is PolynomialOnePositionManipulator) {
          ((PolynomialOnePositionManipulator) op).ContiguityParameter.Value = new DoubleValue(8);
        } else if (op is UniformOnePositionRealVectorManipulator) {
          // nothing to do
        } else {
          um.Operators.SetItemCheckedState(op, false);
        }
      }
      var tm = (MultiIntegerVectorManipulator) TransportTargetManipulator;
      foreach (var op in tm.Operators) {
        if (op is UniformSomePositionsManipulator ||
            op is UniformOnePositionManipulator ||
            op is CommonTargetIntegerVectorMutator ||
            op is CommonTargetSwappingIntegerVectorMutator ||
            op is TargetMergingIntegerVectorManipulator ||
            op is DistanceBasedIntegerVectorMutator) {
          // accept
        } else {
          tm.Operators.SetItemCheckedState(op, false);
        }
      }
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CompoundMutator(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey("MaximumGenerations"))
        Parameters.Add(new LookupParameter<IntValue>("MaximumGenerations", "The maximum number of generations."));
      if (!Parameters.ContainsKey("Generations"))
        Parameters.Add(new LookupParameter<IntValue>("Generations", "The current number of generations."));
      if (!Parameters.ContainsKey("MinGenerationPercentageForUtilzationMutation"))
        Parameters.Add(new ValueLookupParameter<PercentValue>("MinGenerationPercentageForUtilzationMutation", "The minimum percentage of generations (Generations/MaxGenerations) until mutation of utilizations become enabled.", new PercentValue(0.5)));
      if (!Parameters.ContainsKey("ProbabilityOfCombinedMutation"))
        Parameters.Add(new ValueLookupParameter<PercentValue>("ProbabilityOfCombinedMutation", "Probability of mutating both utilization and transport. Don't set to 100% otherwise intermediates are not touched anymore.", new PercentValue(0.3)));
    }
    #endregion

    public override IOperation Apply() {
      var ops = new OperationCollection(base.Apply());
      var targets = ProblemData.TransportTargets.CheckedItems.Select(t => t.Value.Value).ToArray();
      var utilizations = ProblemData.Utilizations.CheckedItems.Select(t => t.Value.Value).ToArray();
      var feedstockUtilizationManipulator = FeedstockUtilizationManipulator;
      if (1.0*Generations/MaximumGenerations < MinGenerationPercentageForUtilzationMutation && TransportTargetManipulator != null) {
        feedstockUtilizationManipulator = null;
      }
      if (feedstockUtilizationManipulator == null && TransportTargetManipulator == null) {
        // do nothing
      } else if (feedstockUtilizationManipulator == null) {
        ops.AddRange(CreateTransportMutation(targets[Random.Next(targets.Length)]));
      } else if (TransportTargetManipulator == null) {
        ops.AddRange(CreateUtilizationMutation(utilizations[Random.Next(utilizations.Length)]));
      } else {
        if (Random.Next() > ProbabilityOfCombinedMutation) {
          int n = utilizations.Length;
          int r = Random.Next(n + targets.Length);
          ops.AddRange(
            r < n
              ? CreateUtilizationMutation(utilizations[r])
              : CreateTransportMutation(targets[r - n]));
        } else {
          var feedstocks = new HashSet<string>(utilizations.Concat(targets)).ToArray();
          var feedstock = feedstocks[Random.Next(feedstocks.Length)];
          ops.AddRange(CreateUtilizationMutation(feedstock));
          ops.AddRange(CreateTransportMutation(feedstock));
        }
      }
      return ops;
    }

    private IEnumerable<IOperation> CreateUtilizationMutation(string name) {
      yield return ExecutionContext.CreateChildOperation(new FeedstockDeflater {ProductName = name});
      var u = (IRealVectorManipulator) FeedstockUtilizationManipulatorParameter.ActualValue.Clone();
      u.RealVectorParameter.ActualName = LayerDescriptor.Utilizations.NameWithPrefix(name);
      yield return ExecutionContext.CreateChildOperation(u);
      yield return ExecutionContext.CreateChildOperation(new FeedstockInflater {ProductName = name});
    }

    private IEnumerable<IOperation> CreateTransportMutation(string name) {
      yield return ExecutionContext.CreateChildOperation(new FeedstockDeflater {ProductName = name});
      var t = (IIntegerVectorManipulator)TransportTargetManipulator.Clone();
      t.IntegerVectorParameter.ActualName = LayerDescriptor.TransportTargets.NameWithPrefix(name);
      yield return ExecutionContext.CreateChildOperation(t);
      yield return ExecutionContext.CreateChildOperation(new FeedstockInflater {ProductName = name});
    }

  }
}
